/*
 *
 *  bluespam 
 *  sends a file to all OPUSH capable devices in piconet
 *
 *  Copyright (C) 2004  Simon de Bakker <simon@v2.nl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation;
 *
 *  Software distributed under the License is distributed on an "AS
 *  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 *  implied. See the License for the specific language governing
 *  rights and limitations under the License.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <btopush/btopush.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <dirent.h>
#include <time.h>
#include <db.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>

#define BSP_NAME	"bluespam"
#define	BSP_DB_ENV_HOME "/mnt/mmc/db"
#define BSP_DB_NAME	"btaddr.db" 
#define BSP_FILEDIR	"/mnt/mmc/content"

static DB_ENV * dbenv = NULL;
static DB * dbp = NULL;
static struct dirent ** filelist;
static int numfiles;

typedef struct dbentry_t {
    char addr[18];
    uint8_t ch;
    uint8_t r;
    uint8_t c;
} dbentry_t;

#define DBT_SIZE	sizeof(DBT)
#define DBENTRY_SIZE	sizeof(dbentry_t)

static int file_select(const struct dirent * d)
{
    char fpath[128];
    struct stat st;
   
    snprintf(fpath, 127, "%s/%s", BSP_FILEDIR, d->d_name);
    
    if (stat(fpath, &st) < 0) {
	printf("%s: %s\n", fpath, strerror(errno));
	return 0;
    }

    if (!S_ISREG(st.st_mode))
	return 0;
    
    return 1;
}

static int read_dir()
{
    return
	(numfiles = scandir(BSP_FILEDIR, &filelist, file_select, versionsort));
}

static int db_open()
{
    int retval;

    if ((retval = db_env_create(&dbenv, 0)) != 0) {
	fprintf(stderr, "Error creating env handle: %s\n",
		db_strerror(retval));
	return -1;
    }
    
    retval = dbenv->open(dbenv,
			 BSP_DB_ENV_HOME,
			 DB_CREATE|DB_PRIVATE|DB_INIT_MPOOL|DB_INIT_LOG,
			 0664);
    if (retval != 0) {
	fprintf(stderr, "Error opening env: %s\n",
		db_strerror(retval));
	return -1;
    }

    if ((retval = db_create(&dbp, dbenv, 0)) != 0) {
	fprintf(stderr, "Error creating db handle: %s\n",
		db_strerror(retval));
	return -1;
    }
  
    dbp->set_errfile(dbp, stderr);
    dbp->set_errpfx(dbp, BSP_NAME);
  
    retval = dbp->open(dbp,
		       NULL,
		       BSP_DB_NAME,
		       NULL,
		       DB_HASH,
		       DB_CREATE,
		       0664);
   
    if (retval != 0) {
	fprintf(stderr, "Error opening db: %s\n",
		db_strerror(retval));
	return -1;
    }

    return 1;
}

static void db_close()
{
    if (dbp)
	dbp->close(dbp, DB_NOSYNC);
    if (dbenv) {
	/* we sync on every put so no need for sync here 
	 * FIXME remove if using transactions (which we should) */
	dbenv->close(dbenv, 0);
    }
}

static int db_dev_exists(char * addr, uint8_t * r, uint8_t * c)
{
    DBT key, data;
    dbentry_t entry;
    int retval;

    memset(&key, 0, DBT_SIZE);
    memset(&data, 0, DBT_SIZE);
    memset(&entry, 0, DBENTRY_SIZE);

    key.data = addr;
    key.size = strlen(addr) + 1;
    
    data.data = &entry;
    data.ulen = DBENTRY_SIZE;
    data.flags = DB_DBT_USERMEM;

    retval = dbp->get(dbp, NULL, &key, &data, 0);
    if (retval == DB_NOTFOUND) 
	return 0;

    printf("found address: %s\n\tchannel: %d\n\tretry: %d\n\tcycle: %d\n",
	    entry.addr, 
	    entry.ch,
	    entry.r,
	    entry.c);
   
    if (r!=NULL)
	*r = entry.r;
    if (c!=NULL)
	*c = entry.c;

    return 1;
}

static int db_dev_put(btopush_dev_t * dev, uint8_t r, uint8_t c)
{
    DBT key, data;
    dbentry_t entry;
    int retval;

    memset(&key, 0, DBT_SIZE);
    memset(&data, 0, DBT_SIZE);
    memset(&entry, 0, DBENTRY_SIZE);

    ba2str(&dev->addr, entry.addr);
    entry.ch = (uint8_t)dev->channel;
    entry.r = r;
    entry.c = c;
    
    key.data = entry.addr;
    key.size = strlen(entry.addr) + 1;
    
    data.data = &entry;
    data.size = DBENTRY_SIZE;

    retval = dbp->put(dbp, NULL, &key, &data, 0);
    if (retval != 0) {
	fprintf(stderr, "Error add %s to database: %s\n",
		entry.addr, db_strerror(retval));
	return 0;
    }
    /* FIXME use transactions instead of syncs here 
     * as sync is much more expensive. */
    dbp->sync(dbp, 0);

    return 1;
}

static int db_dev_remove(char * addr)
{
    DBT key;
    int retval;

    memset(&key, 0, DBT_SIZE);
    key.data = addr;
    key.size = strlen(addr) + 1;
    
    retval = dbp->del(dbp, NULL, &key, 0);
    if (retval != 0) {
	fprintf(stderr, "Error removing %s to database: %s\n",
		addr, db_strerror(retval));
	return 0;
    }
    dbp->sync(dbp, 0);
    return 1;
}

int main(int argc, char * argv[])
{
    btopush_ctx_t btctx;
    btopush_dev_t devs[BTOPUSH_MAX_DEV];
    char * rmaddr = NULL;
    int devc, i;

    if (argc >= 2) 
	rmaddr = argv[1];

    /* open database */
    db_open();
    
    if (rmaddr != NULL) {
	if (!db_dev_exists(rmaddr, NULL, NULL)) {
	    printf("no such address: %s\n", rmaddr);
	    exit(0);
	}
	printf("removing address: %s\n", rmaddr);
	db_dev_remove(rmaddr);
	db_close();
	exit(0);
    }

    if (read_dir() <= 0) {
	fprintf(stderr, "No content!");
	return 0;
    }
	
    /* initialize bluetooth */
    btopush_init(&btctx);

    if ((devc = btopush_inq_objpush(devs)) <= 0) {
	fprintf(stderr, "could not find objpush capable devices\n");
	exit(1);
    } 

    for (i=0; i<devc; i++) {
	int j;
	int addtodb = 0, noclose = 0;
	char addr[18];
	uint8_t retr = 0, cycl = 0;
	ba2str(&devs[i].addr, addr);
	
	if (db_dev_exists(addr, &retr, &cycl)) {
	    if (!retr) {
	       fprintf(stdout, "%s exists in database (or max retry count reached), skipping\n", addr);
	       continue; 
	    } else {
		/*FIXME: is there update record? */
		db_dev_remove(addr);
		if (cycl) {
		    cycl--;
		    addtodb = 1;
		    noclose = 1;
		    goto close;
		} else {
		    cycl = 3;
		}
	    }
	} else {
	    retr = 3;
	    cycl = 3;
	}
		    
	if (btopush_attach_dev(&btctx, &devs[i]) != BTOPUSH_SUCCESS) {
	    fprintf(stderr, "%s could not set device\n", addr);
	    goto next;
	}    
	
	if (btopush_open(&btctx) != BTOPUSH_SUCCESS) {
	    fprintf(stderr, "%s could not open connection\n", addr);
	    retr--;
	    addtodb = 1;
	    goto next;
	}	

	if (btopush_connect(&btctx, "prijsobject") != BTOPUSH_SUCCESS) {
	    fprintf(stderr, "%s could not connect\n", addr);
	    retr--;
	    addtodb = 1;
	    goto close;
	}
	
	for (j=0; j<numfiles; j++) {
	    char fpath[128];
	    char * fname = filelist[j]->d_name; 
	    snprintf(fpath, 127, "%s/%s", BSP_FILEDIR, filelist[j]->d_name);

	    fprintf(stdout, "%s start sending %s\n", addr, fname);
	    if (btopush_open_file(&btctx, fpath) != BTOPUSH_SUCCESS) {
		fprintf(stderr, "could not open file: %s\n", fname);
		goto disc;
	    }
	    
	    if (btopush_push_stream(&btctx) != BTOPUSH_SUCCESS) {
		if (btctx.req_state == BTOPUSH_REQS_TIMEOUT) {
		    fprintf(stderr, "%s connection timed out\n", addr);
		    btopush_close_file(&btctx);
		    retr--;
		    addtodb = 1;
		    goto disc;
		} else {
		    fprintf(stderr, "%s cancelled\n", addr);
		    btopush_close_file(&btctx);
		    retr--;
		    addtodb = 1;
		    goto disc;
		}
	    } else {
		fprintf(stdout, "%s stream succesfull\n", addr);
		retr = 0;
		cycl = 0;
		addtodb = 1;
	    }
	    btopush_close_file(&btctx);
	}
disc:
	btopush_disconnect(&btctx);
close:
	if (addtodb) {
	    db_dev_put(&devs[i], retr, cycl);
	    printf("%s added to database (%d, %d)\n", addr, retr, cycl);
	} else {
	    printf("%s not added to database\n", addr);
	}
	if (!noclose)
	    btopush_close(&btctx);
next:
	; 
    }

    db_close();
    return 0;
}
